"""
File name:      __basicmath__.py
Written by:     Niranjan Bhujel
Date:           June 29, 2022
Description:    Contains function for different matrix operations, activation functions and layers.
"""


import os
from jinja2 import Environment, FileSystemLoader
import tensorflow as tf
from tensorflow import keras


def matrix_operation(operation, data_type):
    """
    Generates C code for specified matrix operation.

    Parameters
    ----------
    operation : str
        Specified matrix operations. Allowed values are: 'mat_vec', 'dot_prod', 'mat_add', mat_assign'.
    data_type : str
        Data type: Either 'float' or 'double'

    Returns
    -------
    str
        C code in str type. 
    """
    templatePath = f"{os.path.dirname(os.path.realpath(__file__))}/templates"

    file_loader = FileSystemLoader(templatePath)
    env = Environment(loader=file_loader)

    Ops = {
        'mat_vec': 'mat_vec_prod.j2',
        'dot_prod': 'dot_prod.j2',
        'mat_add': 'mat_add.j2',
        'mat_assign': 'mat_assign.j2',
    }

    template = env.get_template(f'/matrix_operations/{Ops[operation]}')
    code_output = template.render(data_type=data_type)

    return code_output


def function_operation(function, data_type):
    """
    Generates C code for function operation (activation function).

    Parameters
    ----------
    function : str
        Activation function name. Allowed values are 'tanh', 'sigmoid', 'relu', 'softmax'.
    data_type : str
        Data type: Either 'float' or 'double'

    Returns
    -------
    str
        C code in str type.
    """
    templatePath = f"{os.path.dirname(os.path.realpath(__file__))}/templates"

    file_loader = FileSystemLoader(templatePath)
    env = Environment(loader=file_loader)

    Ops = {
        'tanh': 'tanh.j2',
        'sigmoid': 'sigmoid.j2',
        'relu': 'relu.j2',
        'softmax': 'softmax.j2',
    }

    template = env.get_template(f'/functions/{Ops[function]}')
    code_output = template.render(data_type=data_type)

    return code_output


def layer(layer_type, layer_name, num_cell, num_cell_prev, input_name, output_name, weights, biases, activation, data_type, tab_sep=''):
    """
    Generates C code for different types of layer.

    Parameters
    ----------
    layer_type : str
        Type of the layer. Allowed value are: 'dense', 'flatten'.
    layer_name : str
        Name of the layer.
    num_cell : int
        Number of cell in current layer.
    num_cell_prev : int
        Number of cell in previous layer.
    input_name : str
        Name of the input variable for the layer.
    output_name : str
        Name of the output variable for the layer.
    weights : numpy.array
        Weight matrix
    biases : numpy.array
        Bias vector
    activation : str
        Name of the activation function.
    data_type : str
        Either 'float' or 'double'
    tab_sep : str
        String to have at the start of each line of C code. Defaults to ''.

    Returns
    -------
    str
        C code for the layer.
    """
    templatePath = f"{os.path.dirname(os.path.realpath(__file__))}/templates"
    file_loader = FileSystemLoader(templatePath)
    env = Environment(loader=file_loader)

    Ops = {
        'dense': 'dense.j2',
    }

    template = env.get_template(f'/layers/{Ops[layer_type]}')

    code_output = template.render(
        layer_name=layer_name,
        num_cell=num_cell,
        num_cell_prev=num_cell_prev,
        inputs=input_name,
        outputs=output_name,
        activation=activation,
        weights=weights.T.flatten(),
        biases=biases,
        data_type=data_type,
        tab_sep=tab_sep,
    )

    return code_output
