"""
File name:      __init__.py
Written by:     Niranjan Bhujel
Date:           June 29, 2022
Description:    Contains function to convert keras model to C code.
"""


import tensorflow as tf
from tensorflow import keras
import prettytable
from tf2c import __basicmath__


def __tf2c_pref__(file_name, data_type, activation_list):
    """
    Generates common part of C code for the model.

    It includes function for different matrix operations and activation funcion.

    Parameters
    ----------
    file_name : str
        File name for code generation.
    data_type : str
        'float' or 'double'
    activation_list : list
        List of activation functions whose C function are to be added i.e., list of activation functions required by the model.

    Returns
    -------
    str
        C code
    """
    activation_all = {
        keras.activations.sigmoid: 'sigmoid',
        keras.activations.tanh: 'tanh',
        keras.activations.relu: 'relu',
        keras.activations.softmax: 'softmax',
    }

    Code = f'#include <math.h>\n'
    Code += f'#include <stdlib.h>\n'
    Code += f'#include "{file_name}.h"\n\n'
    Code += __basicmath__.matrix_operation('mat_vec', data_type) + "\n\n"
    Code += __basicmath__.matrix_operation('mat_assign', data_type) + "\n\n"
    Code += __basicmath__.matrix_operation('mat_add', data_type) + "\n\n"

    for k in range(len(activation_list)):
        Code += __basicmath__.function_operation(
            activation_list[k], data_type) + "\n\n"

    return Code


def __tf2c__one_model__(model, func_name, data_type, activation_list_in=[], memory_alloc='static'):
    """
    Generates C code for one keras model

    Parameters
    ----------
    model : keras.model
        Keras model whose C code is desired
    func_name : str
        Desired function name in generated C code
    data_type : str
        'float' or 'double'
    activation_list_in : list, optional
        List of activation function that are desired in the C code. It automatically adds new activation function if it is not in the provided list, by default []
    memory_alloc : str, optional
        'static' or 'dynamic' memory allocation for array in generated C code, by default 'static'

    Returns
    -------
    tuples of str, str, list
        C code, header file code and list of activation (provided as argument and other activation required by model)
    """
    activation_all = {
        keras.activations.sigmoid: 'sigmoid',
        keras.activations.tanh: 'tanh',
        keras.activations.relu: 'relu',
        keras.activations.softmax: 'softmax',
    }

    activation_list = activation_list_in.copy()
    N_input = model.layers[0].input_shape
    if isinstance(N_input, list):
        N_input = N_input[0]
    if isinstance(N_input, tuple):
        tmp = 1
        for a in N_input:
            if a is not None:
                tmp = tmp*a
    N_input = tmp

    max_cell = N_input
    for k in range(1, len(model.layers)):
        try:
            max_cell = max(max_cell, model.layers[k].get_weights()[1].shape[0])
        except:
            pass

        if activation_all[model.layers[k].activation] not in activation_list:
            activation_list.append(activation_all[model.layers[k].activation])

    num_cell = N_input

    # Code = __tf2c_pref__(file_name, data_type, activation_list)
    Code = ""
    Code += f"void {func_name}({data_type} *x, {data_type} *y)\n"
    Code += "{\n"
    if memory_alloc == 'static':
        Code += f"\t{data_type} WEIGHTS_TEMPORARY[{max_cell*max_cell}];\n"
        Code += f"\t{data_type} BIASES_TEMPORARY[{max_cell}];\n"
        Code += f"\t{data_type} INPUTS_TEMPORARY[{max_cell}];\n"
        Code += f"\t{data_type} OUTPUTS_TEMPORARY[{max_cell}];\n"
    elif memory_alloc == 'dynamic':
        Code += f"\t{data_type} *WEIGHTS_TEMPORARY = ({data_type}*) calloc({max_cell*max_cell}, sizeof({data_type}));\n"
        Code += f"\t{data_type} *BIASES_TEMPORARY = ({data_type}*) calloc({max_cell}, sizeof({data_type}));\n"
        Code += f"\t{data_type} *INPUTS_TEMPORARY = ({data_type}*) calloc({max_cell}, sizeof({data_type}));\n"
        Code += f"\t{data_type} *OUTPUTS_TEMPORARY = ({data_type}*) calloc({max_cell}, sizeof({data_type}));\n\n"
    else:
        raise Exception("memory_alloc can be either static or dynamics!!!")

    Code += f"\tmat_assign(OUTPUTS_TEMPORARY, x, {num_cell});\n"

    for k in range(len(model.layers)):
        if isinstance(model.layers[k], keras.layers.Flatten):
            pass
        elif isinstance(model.layers[k], keras.layers.Dense):
            weights = model.layers[k].get_weights()[0].T
            biases = model.layers[k].get_weights()[1]
            layer_name = ""
            try:
                layer_name = model.layers[k].get_config()['name']
            except:
                pass
            Code += f"\tmat_assign(INPUTS_TEMPORARY, OUTPUTS_TEMPORARY, {weights.shape[1]});\n"
            Code += __basicmath__.layer(
                layer_type='dense',
                layer_name=layer_name,
                num_cell=weights.shape[0],
                num_cell_prev=weights.shape[1],
                input_name='INPUTS_TEMPORARY',
                output_name='OUTPUTS_TEMPORARY',
                weights=weights.flatten(),
                biases=biases,
                activation=activation_all[model.layers[k].activation] + '_func',
                data_type=data_type,
                tab_sep='\t'
            )
        else:
            raise Exception("Layer not supported!!!")
        Code += "\n\n"
    Code += f"\tmat_assign(y, OUTPUTS_TEMPORARY, {weights.shape[0]});\n"

    if memory_alloc == 'dynamic':
        Code += "free(WEIGHTS_TEMPORARY);\n"
        Code += "free(BIASES_TEMPORARY);\n"
        Code += "free(INPUTS_TEMPORARY);\n"
        Code += "free(OUTPUTS_TEMPORARY);\n"

    Code += "}"

    Code_h = f"void {func_name}({data_type} *x, {data_type} *y);\n"

    return Code, Code_h, activation_list


def tf2c(model, func_name, file_name, data_type, memory_alloc='static'):
    """
    Function to generate C code from keras model.   

    Parameters
    ----------
    model : keras.model or list
        keras model or list of keras model for which C code is to be generated.
    func_name : str or list
        Name of desired function. It should be string if keras model is passed as model argument.
        It should be list of string if list of keras models are passed.
    file_name : str
        Desired file name (without extension) of the generated code.
    data_type : str
        'float' or 'double'. Whether float or double datatype is to be used in generated code.
    memory_alloc : str, optional
        Type of memory allocation. Either 'static' or 'dynamic', by default 'static'

    Example
    -------
    >>> import tensorflow as tf
    >>> from tensorflow import keras
    >>> import tf2c
    >>> model = keras.models.Sequential()
    >>> model.add(keras.layers.Flatten(input_shape=(3, )))
    >>> model.add(keras.layers.Dense(5, activation="relu"))
    >>> model.add(keras.layers.Dense(5, activation="relu"))
    >>> model.add(keras.layers.Dense(1, activation="tanh"))
    >>> tf2c.tf2c(
            model=model,
            func_name='model_call',
            file_name='model_file',
            data_type='float',
            memory_alloc='static'
        )
    """
    if not isinstance(model, list):
        C, H, Act_List = __tf2c__one_model__(
            model, func_name, data_type, [], memory_alloc)
        PREF = __tf2c_pref__(file_name, data_type, Act_List)

        Code = PREF + "\n\n" + C
        with open(file_name + '.c', 'w') as fw:
            fw.write(Code)
        with open(file_name + '.h', 'w') as fw:
            fw.write(H)

    elif isinstance(model, list):
        Code = ""
        Code_h = ""
        Activation_List = []

        for k in range(len(model)):
            C, H, Act_List = __tf2c__one_model__(
                model[k], func_name[k], data_type, Activation_List, memory_alloc)
            Code += C + "\n\n"
            Code_h += H + "\n\n"
            Activation_List = Act_List.copy()

        PREF = __tf2c_pref__(file_name, data_type, Activation_List)

        Code = PREF + "\n" + Code
        with open(file_name + '.c', 'w') as fw:
            fw.write(Code)
        with open(file_name + '.h', 'w') as fw:
            fw.write(Code_h)


def get_config(model, keywords):
    """
    Get configuration of keras model in tabular form.

    Parameters
    ----------
    model : keras model
        Model for which configuration is required.
    keywords : list of str
        List of keys. Values for supplied keys are shown. 
        For example: 'name', 'batch_input_shape', 'dtype', 'units', 'activation'

    Returns
    -------
    prettytable
        Table which can be displayed using print() command.

    Example
    -------
    >>> import tensorflow as tf
    >>> from tensorflow import keras
    >>> import tf2c
    >>> model = keras.models.Sequential()
    >>> model.add(keras.layers.Flatten(input_shape=(3, )))
    >>> model.add(keras.layers.Dense(5, activation="relu"))
    >>> model.add(keras.layers.Dense(5, activation="relu"))
    >>> model.add(keras.layers.Dense(1, activation="tanh"))
    >>> tf2c.get_config(model, ['batch_input_shape', 'units', 'activation'])
    +---------+-------------------+-------+------------+
    |   name  | batch_input_shape | units | activation |
    +---------+-------------------+-------+------------+
    | flatten |     (None, 3)     |       |            |
    |  dense  |                   |   5   |    relu    |
    | dense_1 |                   |   5   |    relu    |
    | dense_2 |                   |   1   |    tanh    |
    +---------+-------------------+-------+------------+
    """
    myTable = prettytable.PrettyTable(['name'] + keywords)
    for k in range(len(model.layers)):
        data = model.layers[k].get_config()
        tmp = []
        if 'name' in list(data.keys()):
            tmp.append(data['name'])
        else:
            tmp.append('unknown')

        for i in range(len(keywords)):
            try:
                tmp.append(data[keywords[i]])
            except:
                tmp.append("")
        myTable.add_row(tmp)

    return myTable


if __name__ == "__main__":
    model = keras.models.Sequential()
    model.add(keras.layers.Flatten(input_shape=[5, ]))
    model.add(keras.layers.Dense(4, activation="relu"))
    model.add(keras.layers.Dense(3, activation="relu"))
    model.add(keras.layers.Dense(2, activation="softmax"))

    tf2c([model], ['test1'], 'testfile', 'float')
